@@ -5,8 +5,8 @@ extern void init_log(char *progname);
 extern void logp(const char *fmt, ...);
 extern void logc(const char *fmt, ...);
 extern const char *progname(void);
-extern int set_logfp(FILE *fp);
+extern int set_logfp(FILE *fp, struct config *conf);
 extern FILE *get_logfp(void);
-extern int open_logfile(const char *logfile);
+extern int open_logfile(const char *logfile, struct config *conf);
 
 #endif
