@@ -15,7 +15,7 @@
 #include <librsync.h>
 
 // Also used by backup_phase4_server.c
-int do_patch(const char *dst, const char *del, const char *upd, bool gzupd, struct cntr *cntr, struct config *cconf)
+int do_patch(const char *dst, const char *del, const char *upd, bool gzupd, int compression, struct cntr *cntr, struct config *cconf)
 {
 	FILE *dstp=NULL;
 	FILE *delfp=NULL;
@@ -32,7 +32,7 @@ int do_patch(const char *dst, const char *del, const char *upd, bool gzupd, stru
 		return -1;
 	}
 
-	if(dpth_is_compressed(del))
+	if(dpth_is_compressed(compression, del))
 		delzp=gzopen(del, "rb");
 	else
 		delfp=fopen(del, "rb");
@@ -69,7 +69,7 @@ int do_patch(const char *dst, const char *del, const char *upd, bool gzupd, stru
 	return result;
 }
 
-static int inflate_or_link_oldfile(const char *oldpath, const char *infpath)
+static int inflate_or_link_oldfile(const char *oldpath, const char *infpath, int compression)
 {
 	int ret=0;
 	struct stat statp;
@@ -80,7 +80,7 @@ static int inflate_or_link_oldfile(const char *oldpath, const char *infpath)
 		return -1;
 	}
 
-	if(dpth_is_compressed(oldpath))
+	if(dpth_is_compressed(compression, oldpath))
 	{
 		FILE *source=NULL;
 		FILE *dest=NULL;
@@ -128,7 +128,7 @@ static int inflate_or_link_oldfile(const char *oldpath, const char *infpath)
 	return ret;
 }
 
-static int send_file(const char *fname, int patches, const char *best, const char *datapth, unsigned long long *bytes, char cmd, int64_t winattr, struct cntr *cntr, struct config *cconf)
+static int send_file(const char *fname, int patches, const char *best, const char *datapth, unsigned long long *bytes, char cmd, int64_t winattr, int compression, struct cntr *cntr, struct config *cconf)
 {
 	int ret=0;
 	FILE *fp=NULL;
@@ -159,7 +159,7 @@ static int send_file(const char *fname, int patches, const char *best, const cha
 		// It might have been stored uncompressed. Gzip it during
 		// the send. If the client knew what kind of file it would be
 		// receiving, this step could disappear.
-		else if(!dpth_is_compressed(datapth))
+		else if(!dpth_is_compressed(compression, datapth))
 		{
 			ret=send_whole_file_gz(best, datapth, 1, bytes,
 				NULL, cntr, 9, NULL, fp, NULL, 0);
@@ -176,7 +176,7 @@ static int send_file(const char *fname, int patches, const char *best, const cha
 	return ret;
 }
 
-static int verify_file(const char *fname, int patches, const char *best, const char *datapth, unsigned long long *bytes, const char *endfile, char cmd, struct cntr *cntr)
+static int verify_file(const char *fname, int patches, const char *best, const char *datapth, unsigned long long *bytes, const char *endfile, char cmd, int compression, struct cntr *cntr)
 {
 	MD5_CTX md5;
 	size_t b=0;
@@ -198,7 +198,7 @@ static int verify_file(const char *fname, int patches, const char *best, const c
 	}
 	if(patches
 	  || cmd==CMD_ENC_FILE || cmd==CMD_ENC_METADATA || cmd==CMD_EFS_FILE
-	  || (!patches && !dpth_is_compressed(best)))
+	  || (!patches && !dpth_is_compressed(compression, best)))
 	{
 		// If we did some patches or encryption, or the compression
 		// was turned off, the resulting file is not gzipped.
@@ -275,7 +275,7 @@ static int verify_file(const char *fname, int patches, const char *best, const c
 
 // a = length of struct bu array
 // i = position to restore from
-static int restore_file(struct bu *arr, int a, int i, const char *datapth, const char *fname, const char *tmppath1, const char *tmppath2, int act, const char *endfile, char cmd, int64_t winattr, struct cntr *cntr, struct config *cconf)
+static int restore_file(struct bu *arr, int a, int i, const char *datapth, const char *fname, const char *tmppath1, const char *tmppath2, int act, const char *endfile, char cmd, int64_t winattr, int compression, struct cntr *cntr, struct config *cconf)
 {
 	int x=0;
 	char msg[256]="";
@@ -330,7 +330,8 @@ static int restore_file(struct bu *arr, int a, int i, const char *datapth, const
 				if(!patches)
 				{
 					// Need to gunzip the first one.
-					if(inflate_or_link_oldfile(best, tmp))
+					if(inflate_or_link_oldfile(best, tmp,
+						compression))
 					{
 						logp("error when inflating %s\n", best);
 						free(path);
@@ -343,8 +344,9 @@ static int restore_file(struct bu *arr, int a, int i, const char *datapth, const
 				}
 
 				if(do_patch(best, dpath, tmp,
-				  FALSE /* do not gzip the result */, cntr,
-				  cconf))
+				  FALSE /* do not gzip the result */,
+				  compression /* from the manifest */,
+				  cntr, cconf))
 				{
 					char msg[256]="";
 					snprintf(msg, sizeof(msg),
@@ -367,7 +369,8 @@ static int restore_file(struct bu *arr, int a, int i, const char *datapth, const
 			if(act==ACTION_RESTORE)
 			{
 				if(send_file(fname, patches, best, datapth,
-					&bytes, cmd, winattr, cntr, cconf))
+					&bytes, cmd, winattr, compression,
+					cntr, cconf))
 				{
 					free(path);
 					return -1;
@@ -377,7 +380,8 @@ static int restore_file(struct bu *arr, int a, int i, const char *datapth, const
 			else if(act==ACTION_VERIFY)
 			{
 				if(verify_file(fname, patches, best, datapth,
-					&bytes, endfile, cmd, cntr))
+					&bytes, endfile, cmd, compression,
+					cntr))
 				{
 					free(path);
 					return -1;
@@ -411,7 +415,8 @@ static int restore_sbuf(struct sbuf *sb, struct bu *arr, int a, int i, const cha
 	{
 		return restore_file(arr, a, i, sb->datapth,
 		  sb->path, tmppath1, tmppath2, act,
-		  sb->endfile, sb->cmd, sb->winattr, cntr, cconf);
+		  sb->endfile, sb->cmd, sb->winattr,
+		  sb->compression, cntr, cconf);
 	}
 	else
 	{
@@ -473,9 +478,87 @@ static int do_restore_end(enum action act, struct cntr *cntr)
 	return ret;
 }
 
+static int restore_ent(const char *client, struct sbuf *sb, struct sbuf ***sblist, int *scount, struct bu *arr, int a, int i, const char *tmppath1, const char *tmppath2, enum action act, char status, struct config *cconf, struct cntr *cntr, struct cntr *p1cntr)
+{
+	int s=0;
+	int ret=0;
+	// Check if we have any directories waiting to be restored.
+	for(s=(*scount)-1; s>=0; s--)
+	{
+		if(is_subdir((*sblist)[s]->path, sb->path))
+		{
+			// We are still in a subdir.
+			//printf(" subdir (%s %s)\n",
+			// (*sblist)[s]->path, sb->path);
+			break;
+		}
+		else
+		{
+			// Can now restore sblist[s] because nothing else is
+			// fiddling in a subdirectory.
+			if(restore_sbuf((*sblist)[s], arr, a, i, tmppath1,
+				tmppath2, act, client, status,
+				p1cntr, cntr, cconf))
+			{
+				ret=-1;
+				break;
+			}
+			else if(del_from_sbuf_arr(sblist, scount))
+			{
+				ret=-1;
+				break;
+			}
+		}
+	}
+
+	/* If it is a directory, need to remember it and restore it later, so
+	   that the permissions come out right. */
+	/* Meta data of directories will also have the stat stuff set to be a
+	   directory, so will also come out at the end. */
+	if(!ret && S_ISDIR(sb->statp.st_mode))
+	{
+		if(add_to_sbuf_arr(sblist, sb, scount))
+			ret=-1;
+
+		// Wipe out sb, without freeing up all the strings inside it,
+		// which have been added to sblist.
+		init_sbuf(sb);
+	}
+	else if(!ret && restore_sbuf(sb, arr, a, i, tmppath1, tmppath2, act,
+		client, status, p1cntr, cntr, cconf))
+			ret=-1;
+	return ret;
+}
+
+static int srestore_matches(struct strlist *s, const char *path)
+{
+	int r=0;
+	if(!s->flag) return 0; // Do not know how to do excludes yet.
+	if((r=strncmp(s->path, path, strlen(s->path)))) return 0; // no match
+	if(!r) return 1; // exact match
+	if(*(path+strlen(s->path)+1)=='/')
+		return 1; // matched directory contents
+	return 0; // no match
+}
+
+/* Used when restore is initiated from the server. */
+static int check_srestore(struct config *cconf, const char *path)
+{
+	int i=0;
+	for(i=0; i<cconf->iecount; i++)
+	{
+		//printf(" %d %s %s\n",
+		//	cconf->incexcdir[i]->flag, cconf->incexcdir[i]->path,
+		//	path);
+		if(srestore_matches(cconf->incexcdir[i], path))
+			return 1;
+	}
+	return 0;
+}
+
 // a = length of struct bu array
 // i = position to restore from
-static int restore_manifest(struct bu *arr, int a, int i, const char *tmppath1, const char *tmppath2, regex_t *regex, enum action act, const char *client, struct cntr *p1cntr, struct cntr *cntr, struct config *cconf, bool all)
+static int restore_manifest(struct bu *arr, int a, int i, const char *tmppath1, const char *tmppath2, regex_t *regex, int srestore, enum action act, const char *client, struct cntr *p1cntr, struct cntr *cntr, struct config *cconf, bool all)
 {
 	int ret=0;
 	gzFile zp=NULL;
@@ -500,7 +583,7 @@ static int restore_manifest(struct bu *arr, int a, int i, const char *tmppath1,
 		log_and_send("out of memory");
 		ret=-1;
 	}
-	else if(!(logfp=open_file(logpath, "ab")) || set_logfp(logfp))
+	else if(!(logfp=open_file(logpath, "ab")) || set_logfp(logfp, cconf))
 	{
 		char msg[256]="";
 		snprintf(msg, sizeof(msg),
@@ -508,7 +591,10 @@ static int restore_manifest(struct bu *arr, int a, int i, const char *tmppath1,
 		log_and_send(msg);
 		ret=-1;
 	}
-	else if(!(zp=gzopen_file(manifest, "rb")))
+
+	log_restore_settings(cconf, srestore);
+
+	if(!ret && !(zp=gzopen_file(manifest, "rb")))
 	{
 		log_and_send("could not open manifest");
 		ret=-1;
@@ -516,12 +602,12 @@ static int restore_manifest(struct bu *arr, int a, int i, const char *tmppath1,
 	else
 	{
 		char cmd;
+		int s=0;
 		int quit=0;
 		size_t len=0;
 		struct sbuf sb;
 		// For out-of-sequence directory restoring so that the
 		// timestamps come out right:
-		int s=0;
 		int scount=0;
 		struct sbuf **sblist=NULL;
 
@@ -571,64 +657,17 @@ static int restore_manifest(struct bu *arr, int a, int i, const char *tmppath1,
 			}
 			else
 			{
-				if(check_regex(regex, sb.path))
+				if((!srestore
+				    || check_srestore(cconf, sb.path))
+				  && check_regex(regex, sb.path)
+				  && restore_ent(client,
+					&sb, &sblist, &scount,
+					arr, a, i, tmppath1, tmppath2,
+					act, status, cconf,
+					cntr, p1cntr))
 				{
-				  // Check if we have any directories waiting
-				  // to be restored.
-				  for(s=scount-1; s>=0; s--)
-				  {
-					if(is_subdir(sblist[s]->path, sb.path))
-					{
-						// We are still in a subdir.
-						//printf(" subdir (%s %s)\n", sblist[s]->path, sb.path);
-						break;
-					}
-					else
-					{
-						// Can now restore sblist[s]
-						// because nothing else is
-						// fiddling in a subdirectory.
-				  		if(restore_sbuf(sblist[s], arr,
-						 a, i, tmppath1, tmppath2, act,
-						 client, status,
-						 p1cntr, cntr, cconf))
-						{
-							ret=-1; quit++;
-							break;
-						}
-						else if(del_from_sbuf_arr(
-							&sblist, &scount))
-						{
-							ret=-1; quit++;
-							break;
-						}
-					}
-				  }
-
-				  /* If it is a directory, need to remember it
-				     and restore it later, so that the
-				     permissions come out right. */
-				  /* Meta data of directories will also have
-				     the stat stuff set to be a directory,
-				     so will also come out at the end. */
-				  if(!ret && S_ISDIR(sb.statp.st_mode))
-				  {
-					if(add_to_sbuf_arr(&sblist, &sb, &scount))
-					{
-						ret=-1; quit++;
-					}
-
-					// Wipe out sb, without freeing up
-					// all the strings inside it, which
-					// have been added to sblist.
-					init_sbuf(&sb);
-				  }
-				  else if(!ret && restore_sbuf(&sb, arr, a, i,
-				    tmppath1, tmppath2, act, client, status,
-				    p1cntr, cntr, cconf))
-				  {
-					ret=-1; quit++;
-				  }
+					ret=-1;
+					quit++;
 				}
 			}
 			free_sbuf(&sb);
@@ -655,7 +694,7 @@ static int restore_manifest(struct bu *arr, int a, int i, const char *tmppath1,
 		reset_filecounter(p1cntr);
 		reset_filecounter(cntr);
 	}
-	set_logfp(NULL);
+	set_logfp(NULL, cconf);
 	compress_file(logpath, logpathz, cconf);
 	if(manifest) free(manifest);
 	if(datadir) free(datadir);
@@ -664,7 +703,7 @@ static int restore_manifest(struct bu *arr, int a, int i, const char *tmppath1,
 	return ret;
 }
 
-int do_restore_server(const char *basedir, const char *backup, const char *restoreregex, enum action act, const char *client, struct cntr *p1cntr, struct cntr *cntr, struct config *cconf)
+int do_restore_server(const char *basedir, enum action act, const char *client, int srestore, struct cntr *p1cntr, struct cntr *cntr, struct config *cconf)
 {
 	int a=0;
 	int i=0;
@@ -679,7 +718,7 @@ int do_restore_server(const char *basedir, const char *backup, const char *resto
 
 	logp("in do_restore\n");
 
-	if(compile_regex(&regex, restoreregex)) return -1;
+	if(compile_regex(&regex, cconf->regex)) return -1;
 
 	if(!(tmppath1=prepend_s(basedir, "tmp1", strlen("tmp1")))
 	  || !(tmppath2=prepend_s(basedir, "tmp2", strlen("tmp2"))))
@@ -697,28 +736,29 @@ int do_restore_server(const char *basedir, const char *backup, const char *resto
 		return -1;
 	}
 
-	if(backup && *backup=='a')
+	if(cconf->backup && *(cconf->backup)=='a')
 	{
 		all=TRUE;
 	}
-	else if(!(index=strtoul(backup, NULL, 10)) && a>0)
+	else if(!(index=strtoul(cconf->backup, NULL, 10)) && a>0)
 	{
 		// No backup specified, do the most recent.
 		ret=restore_manifest(arr, a, a-1,
-			tmppath1, tmppath2, regex, act, client,
+			tmppath1, tmppath2, regex, srestore, act, client,
 			p1cntr, cntr, cconf, all);
 		found=TRUE;
 	}
 
 	if(!found) for(i=0; i<a; i++)
 	{
-		if(all || !strcmp(arr[i].timestamp, backup)
+		if(all || !strcmp(arr[i].timestamp, cconf->backup)
 			|| arr[i].index==index)
 		{
 			found=TRUE;
 			//logp("got: %s\n", arr[i].path);
 			ret|=restore_manifest(arr, a, i,
-				tmppath1, tmppath2, regex, act, client,
+				tmppath1, tmppath2, regex,
+				srestore, act, client,
 				p1cntr, cntr, cconf, all);
 			if(!all) break;
 		}
